import numpy as np
import pickle
import datetime
import copy

from sklearn.preprocessing import LabelEncoder

from fixout import fairness, interface
from fixout.helper import ReverseFairness, UnfairModel, clazzes
from fixout.interface.ttypes import SensitiveFeature, FairMetricEnum

import warnings
warnings.filterwarnings('ignore')

import fixout.web.webapp as interface

from IPython.display import IFrame
import threading

httpaddress = "http://localhost:5000/"

class FixOutRunner:
    """
    A class to process and evaluate fairness in machine learning models, given a FixOut artifact.

    See Also
    --------
    fixout.artifact.FixOutArtifact : A FixOut artifact.
    """
    
    def __init__(self,report_name=""):
        """
        Initializes the FixOutRunner with a report name.
        
        Parameters:
        -----------
        report_name : str, optional
            Name of the report. Defaults to an empty string.
        """
        self.input = {}
        self.input["report_details"] = {}
        self.input["report_details"]["report_name"] = report_name
        self.input["report_details"]["generated"] = datetime.datetime.now().date()

        self.output = {}

    def _common(self, fxa):
         
        self.input["model"] = fxa.model
        self.input["X"] = fxa.X
        self.input["y"] = fxa.y
        self.input["f_names"] = fxa.features_name
        self.input["nonnumeric_features"] = fxa.nonnumeric_features

        self.input["testing_data"] = fxa.test_data
        
        self.input["dictionary"] = fxa.dictionary 

        if self.input["model"] is None and fxa.y_pred is None:
            raise

        if fxa.y_pred is None:
            self.output["y_pred"] = self.input["model"].predict(self.input["X"])
        else:
            self.output["y_pred"] = fxa.y_pred
        self.prob_y_pred = fxa.prob_y_pred


        if len(fxa.sensfeatList) > 0 and type(fxa.sensfeatList[0]) is int: 
            sens_f_indexes = fxa.sensfeatList #[u for u,_,_ in fxa.sensfeatList]
            self.input["sens_f_names"] = [self.input["f_names"][u] for u in fxa.sensfeatList]
        elif len(fxa.sensfeatList) > 0 and type(fxa.sensfeatList[0]) is str:
            sens_f_indexes = [self.input["f_names"].index(w) for w in fxa.sensfeatList]
            self.input["sens_f_names"] = fxa.sensfeatList #[w for _,_,w in fxa.sensfeatList
        else:
            sens_f_indexes = []
            self.input["sens_f_names"] = []

        #sens_f_unprivPops = [v for _,v,_ in fxa.sensfeatList]
        #sens_f_unprivPops_discretes = []
        #self.input["sens_f_names"] = [w for _,w in fxa.sensfeatList] #[w for _,_,w in fxa.sensfeatList]

        encoders = []

        transformed_data = copy.deepcopy(self.input["X"])
        
        for i in range(len(self.input["f_names"])):
            
            le = None
            
            if i in self.input["nonnumeric_features"]:
                le = LabelEncoder( )
                le.fit(self.input["X"][:,i])
                transformed_data[:,i] = le.transform(self.input["X"][:,i]).astype(float)

            encoders.append(le)

        self.input["sens_f_index"] = sens_f_indexes
        
        '''
        ######
        # for each column
        for i in range(len(self.input["sens_f_index"])):
            
            sens_f_index = self.input["sens_f_index"][i]

            if sens_f_index in self.input["nonnumeric_features"]: 

                le = encoders[sens_f_index]
                #sens_f_unprivPops_discreted = int(le.transform([sens_f_unprivPops[i]])[0])
                
                #new_array = [1 if x == str(float(sens_f_unprivPops_discreted)) else 0 for x in transformed_data[:,sens_f_index]]
                #transformed_data[:,sens_f_index] = np.array(new_array)
            
            else:
                sens_f_unprivPops_discreted = int(sens_f_unprivPops[i])
                    
            #sens_f_unprivPops_discretes.append(sens_f_unprivPops_discreted)
        '''
        
        self.sensitivefeatureslist = []
        
        # for each sensitive feature
        for i in range(len(self.input["sens_f_index"])):

            aSensitiveFeature = SensitiveFeature()
            aSensitiveFeature.featureIndex = self.input["sens_f_index"][i] 
            #aSensitiveFeature.unprivPop = None #sens_f_unprivPops_discretes[i]
            #aSensitiveFeature.unprivPop_original = None #sens_f_unprivPops[i]
            aSensitiveFeature.name = self.input["sens_f_names"][i]
            aSensitiveFeature.description = ""
            aSensitiveFeature.type = 1 if self.input["sens_f_index"][i] in self.input["nonnumeric_features"] else 0
            self.sensitivefeatureslist.append(aSensitiveFeature)
        
        ######

        transformed_data = transformed_data.astype(float)
        self.input["X"] = transformed_data
        
        self.input["model_availability"] = self.input["model"] is not None
        self.input["sens_f_unpriv"] = [x.unprivPop for x in self.sensitivefeatureslist],
        self.input["sens_f_unpriv_original"] = [x.unprivPop_original for x in self.sensitivefeatureslist],
        self.input["sens_f_type"] = [1 if x in self.input["nonnumeric_features"] else 0 for x in self.sensitivefeatureslist],
        self.input["sens_f_pair"] = [(x.featureIndex, x.name) for x in self.sensitivefeatureslist]
        
        self.output["prob_y_pred"] = None, # Fix it
        
        rev_fairness = ReverseFairness()
        self.input["reversed_models"] = []
        rev_train = rev_fairness.build_reversed_models(self.input["X"], self.input["y"], self.sensitivefeatureslist)
        self.input["reversed_models"].append(rev_train)
        for X_test,y_test,_ in self.input["testing_data"]:
            rev_test = rev_fairness.build_reversed_models(self.input["X"], self.input["y"], self.sensitivefeatureslist, X_test, y_test)
            self.input["reversed_models"].append(rev_test)
    
        unfair_model = UnfairModel()
        self.input["unfair_model"] = []
        unfair_train =  unfair_model.build_unfair_model(self.input["X"], self.input["y"], self.sensitivefeatureslist)
        self.input["unfair_model"].append(unfair_train)
        for X_test,y_test,_ in self.input["testing_data"]:
            runfair_test = unfair_model.build_unfair_model(self.input["X"], self.input["y"], self.sensitivefeatureslist, X_test, y_test)
            self.input["unfair_model"].append(runfair_test)


        self._assess_fairness()


    def _assess_fairness(self):

        self.output["metrics_list"] = [FairMetricEnum.DP, FairMetricEnum.EO, FairMetricEnum.PE, FairMetricEnum.EOD]
        self.output["nonStandardMetricsToBeCalculated"] = [FairMetricEnum.PP, FairMetricEnum.CEA]

        self.output["result"] = self._eval_fairness(self.output["metrics_list"],
                                                   self.sensitivefeatureslist,
                                                   self.input["X"].tolist(),
                                                   self.input["y"].tolist(),
                                                   self.output["y_pred"],
                                                   "original")
        
        self.output["nonstandardResults"] = self._eval_fairness(self.output["nonStandardMetricsToBeCalculated"],
                                                               self.sensitivefeatureslist,
                                                               self.input["X"].tolist(),
                                                               self.input["y"].tolist(),
                                                               self.output["y_pred"],
                                                               "original")
        self._baselines()

    def _eval_fairness(self,metrics,sensFeatures,X,y,y_pred,txtIndicator):
        
        results = []
        for sensitiveFeature in sensFeatures:
            r = fairness.computeFairnessMetrics(metrics,
                                       sensitiveFeature, 
                                       X, 
                                       y,
                                       y_pred)
            results.append((sensitiveFeature,r,txtIndicator))
        
        return results
    

    def _run(self, fxa, webserver):
        self._common(fxa)
        if webserver :
            pickle.dump((self.input, self.output),open(str("repport_output.fixout"),"wb"))
                

    def run(self, fxa, show=True):
        """
        Runs FixOut with a given artifact.

        Parameters:
        -----------
        fxa : FixOutArtifact
            Original model, training and/or testing data to process.
        show : bool, optional
            If True (default), the output will be shown using a web interface, 
            otherwise only the evaluation will be executed and the results returned.

        Returns:
        --------
        tuple or None
            If `show` is True, returns None and actives a web interface.
            Otherwise, returns the computed evaluation results.

        """
        self._run(fxa, show)
        if show :
            print("Initializing the web interface.\nResults available at http://localhost:5000")
            interface.app.run()
        return self.output
    
    def runJ(self, fxa, show=True):
        """
        Runs FixOut with a given artifact in a Jupyter environment.

        Parameters:
        -----------
        fxa : FixOutArtifact
            Original model, training and/or testing data to process.
        show : bool, optional
            If True (default), the output will be shown using a web interface, 
            otherwise None is returned.

        Returns:
        --------
        IFrame
            Displays a frame with all the results obtained using FixOut.
        """
        self._run(fxa, True)
        self.wserver = threading.Thread(target=interface.app.run,args=())
        self.wserver.start()
        if show:
            return IFrame(httpaddress, 1200,800)
        else:
            return

    
    #def close(self):
    #    self.wserver.join()
    #    return 
    
    def _buildURL(self,service_name,dataset_reference):
        address = httpaddress + "d/" + service_name + "/"
        
        if dataset_reference is not None: # checking testing data
            if type(dataset_reference) is int:
                return address + str(dataset_reference+1)
            else:
                # searching for index based on label dataset
                for i in range(len(self.input["testing_data"])):
                    _,_,label = self.input["testing_data"][i] # X, y, label
                    if label == dataset_reference:
                        return address + str(i+1)
                return None # not found and is not training data
        else: # training data
            return address + str(0)
    
    def show(self):
        return IFrame(httpaddress, 1200,800)

    def data_distribution(self, dataset_reference=None):
        """
        Shows histograms to analyze the distribution of data centered on sensitive features.

        Returns:
        --------
        IFrame
            A frame with the requested results 
        """
        address = self._buildURL("histo",dataset_reference)
        if address is None:
            return 
        return IFrame(address, 1200,400)
    
    def data_visualization(self, dataset_reference=None):
        """
        Shows histograms to analyze the distribution of data centered on sensitive features.

        Returns:
        --------
        IFrame
            A frame with the requested results 
        """
        address = self._buildURL("visu",dataset_reference)
        if address is None:
            return 
        return IFrame(address, 1200,500)
    
    def correlation(self, dataset_reference=None):
        """
        Shows correlation matrices and rankings of features ordered according to a correlation coefficient. 
        The rankings are generated centered on sensitive features.

        Returns:
        --------
        IFrame
            A frame with the requested results 
        """
        address = self._buildURL("corr",dataset_reference)
        if address is None:
            return 
        return IFrame(address, 1200,400)
    
    def reverse(self, dataset_reference=None):
        """
        Shows the performance metrics for reversed models. 
        These models attempt to predict sensitive features from target labels.

        Returns:
        --------
        IFrame
            A frame with the requested results 
        """
        address = self._buildURL("rev",dataset_reference)
        if address is None:
            return 
        return IFrame(address, 1200,400)
    
    def discriminatory(self, dataset_reference=None):
        address = self._buildURL("corr",dataset_reference)
        if address is None:
            return 
        return IFrame(address, 1200,800)
    
    def fairness(self,dataset_reference=None):
        """
        Shows the calculated fairness metrics for the provided model. 

        Returns:
        --------
        IFrame
            A frame with the requested results 
        """
        address = self._buildURL("metrics",dataset_reference)
        return IFrame("http://localhost:5000/d/metrics/1", 1200,800)


    def get_fairness(self,
                     model="original",
                     sensitivefeature=None):
        """
        Retrieves calculated fairness metrics for a given model and sensitive feature.

        Parameters:
        -----------
        model : str, optional
            The model label to filter results (default: "original").
        sensitivefeature : str, optional
            The specific sensitive feature to filter results (default: None).

        Returns:
        --------
        dict
            A dictionary where keys are model labels and values are the calculated fairness metrics.
        """
        
        result = {}

        for sensf, calculated_metrics, model_label in self.output["result"]:#, self.output["nonstandardResults"]]:
            if sensitivefeature is not None and sensitivefeature == sensf.name:
                result[model_label] = calculated_metrics
            elif sensitivefeature is None:
                result[model_label] = calculated_metrics

        return result        

    def _baselines(self):

        predictions_list=[]

        for clazz,clazz_name in clazzes:
            self.__build_model(clazz,clazz_name,predictions_list)

        for name_method, preditions in predictions_list:
            for sensitiveFeature in self.sensitivefeatureslist:
                r = fairness.computeFairnessMetrics(self.output["metrics_list"],
                                        sensitiveFeature, 
                                        self.input["X"].tolist(), 
                                        self.input["y"].tolist(),
                                        preditions)
                self.output["result"].append((sensitiveFeature,r,name_method))

                nonStandardR = fairness.computeFairnessMetrics(self.output["nonStandardMetricsToBeCalculated"],
                                       sensitiveFeature, 
                                       self.input["X"].tolist(), 
                                       self.input["y"].tolist(),
                                       preditions)
                self.output["nonstandardResults"].append((sensitiveFeature,nonStandardR,name_method))

    def __build_model(self, clazz, name_method, predictions_list):

        clf = clazz()
        clf.fit(self.input["X"], self.input["y"])
        y_pred = clf.predict(self.input["X"])
        predictions_list.append((name_method, y_pred))




#
# Autogenerated by Thrift Compiler (0.18.1)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class ModelType(object):
    LR = 0
    RF = 1
    ADA = 2
    BAG = 3

    _VALUES_TO_NAMES = {
        0: "LR",
        1: "RF",
        2: "ADA",
        3: "BAG",
    }

    _NAMES_TO_VALUES = {
        "LR": 0,
        "RF": 1,
        "ADA": 2,
        "BAG": 3,
    }


class EvalMethod(object):
    EXP = 0
    MET = 1
    BOTH = 2

    _VALUES_TO_NAMES = {
        0: "EXP",
        1: "MET",
        2: "BOTH",
    }

    _NAMES_TO_VALUES = {
        "EXP": 0,
        "MET": 1,
        "BOTH": 2,
    }


class FairMetricEnum(object):
    DP = 0
    EO = 1
    PE = 2
    EOD = 3
    PP = 4
    CEA = 5

    _VALUES_TO_NAMES = {
        0: "DP",
        1: "EO",
        2: "PE",
        3: "EOD",
        4: "PP",
        5: "CEA",
    }

    _NAMES_TO_VALUES = {
        "DP": 0,
        "EO": 1,
        "PE": 2,
        "EOD": 3,
        "PP": 4,
        "CEA": 5,
    }


class FairMetric(object):
    """
    Attributes:
     - type
     - value

    """


    def __init__(self, type=None, value=None,):
        self.type = type
        self.value = value

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.type = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.DOUBLE:
                    self.value = iprot.readDouble()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('FairMetric')
        if self.type is not None:
            oprot.writeFieldBegin('type', TType.I32, 1)
            oprot.writeI32(self.type)
            oprot.writeFieldEnd()
        if self.value is not None:
            oprot.writeFieldBegin('value', TType.DOUBLE, 2)
            oprot.writeDouble(self.value)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class SensitiveFeature(object):
    """
    Attributes:
     - featureIndex
     - name
     - description
     - unprivPop
     - unprivPop_original
     - type

    """


    def __init__(self, featureIndex=None, name=None, description=None, unprivPop=None, unprivPop_original=None, type=None,):
        self.featureIndex = featureIndex
        self.name = name
        self.description = description
        self.unprivPop = unprivPop
        self.unprivPop_original = unprivPop_original
        self.type = type

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.featureIndex = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.name = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.STRING:
                    self.description = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.I32:
                    self.unprivPop = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.STRING:
                    self.unprivPop_original = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 6:
                if ftype == TType.I32:
                    self.type = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('SensitiveFeature')
        if self.featureIndex is not None:
            oprot.writeFieldBegin('featureIndex', TType.I32, 1)
            oprot.writeI32(self.featureIndex)
            oprot.writeFieldEnd()
        if self.name is not None:
            oprot.writeFieldBegin('name', TType.STRING, 2)
            oprot.writeString(self.name.encode('utf-8') if sys.version_info[0] == 2 else self.name)
            oprot.writeFieldEnd()
        if self.description is not None:
            oprot.writeFieldBegin('description', TType.STRING, 3)
            oprot.writeString(self.description.encode('utf-8') if sys.version_info[0] == 2 else self.description)
            oprot.writeFieldEnd()
        if self.unprivPop is not None:
            oprot.writeFieldBegin('unprivPop', TType.I32, 4)
            oprot.writeI32(self.unprivPop)
            oprot.writeFieldEnd()
        if self.unprivPop_original is not None:
            oprot.writeFieldBegin('unprivPop_original', TType.STRING, 5)
            oprot.writeString(self.unprivPop_original.encode('utf-8') if sys.version_info[0] == 2 else self.unprivPop_original)
            oprot.writeFieldEnd()
        if self.type is not None:
            oprot.writeFieldBegin('type', TType.I32, 6)
            oprot.writeI32(self.type)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class FeatureImportance(object):
    """
    Attributes:
     - indexFeature
     - importance

    """


    def __init__(self, indexFeature=None, importance=None,):
        self.indexFeature = indexFeature
        self.importance = importance

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.indexFeature = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.DOUBLE:
                    self.importance = iprot.readDouble()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('FeatureImportance')
        if self.indexFeature is not None:
            oprot.writeFieldBegin('indexFeature', TType.I32, 1)
            oprot.writeI32(self.indexFeature)
            oprot.writeFieldEnd()
        if self.importance is not None:
            oprot.writeFieldBegin('importance', TType.DOUBLE, 2)
            oprot.writeDouble(self.importance)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class FIExplanation(object):
    """
    Attributes:
     - listOfFeatureImportance

    """


    def __init__(self, listOfFeatureImportance=None,):
        self.listOfFeatureImportance = listOfFeatureImportance

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.LIST:
                    self.listOfFeatureImportance = []
                    (_etype3, _size0) = iprot.readListBegin()
                    for _i4 in range(_size0):
                        _elem5 = FeatureImportance()
                        _elem5.read(iprot)
                        self.listOfFeatureImportance.append(_elem5)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('FIExplanation')
        if self.listOfFeatureImportance is not None:
            oprot.writeFieldBegin('listOfFeatureImportance', TType.LIST, 1)
            oprot.writeListBegin(TType.STRUCT, len(self.listOfFeatureImportance))
            for iter6 in self.listOfFeatureImportance:
                iter6.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class ModelDefinitions(object):
    """
    Attributes:
     - numFeatures
     - nonNumericFeatures
     - type

    """


    def __init__(self, numFeatures=None, nonNumericFeatures=None, type=None,):
        self.numFeatures = numFeatures
        self.nonNumericFeatures = nonNumericFeatures
        self.type = type

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.numFeatures = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.nonNumericFeatures = []
                    (_etype10, _size7) = iprot.readListBegin()
                    for _i11 in range(_size7):
                        _elem12 = iprot.readI32()
                        self.nonNumericFeatures.append(_elem12)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I32:
                    self.type = iprot.readI32()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('ModelDefinitions')
        if self.numFeatures is not None:
            oprot.writeFieldBegin('numFeatures', TType.I32, 1)
            oprot.writeI32(self.numFeatures)
            oprot.writeFieldEnd()
        if self.nonNumericFeatures is not None:
            oprot.writeFieldBegin('nonNumericFeatures', TType.LIST, 2)
            oprot.writeListBegin(TType.I32, len(self.nonNumericFeatures))
            for iter13 in self.nonNumericFeatures:
                oprot.writeI32(iter13)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.type is not None:
            oprot.writeFieldBegin('type', TType.I32, 3)
            oprot.writeI32(self.type)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Instance(object):
    """
    Attributes:
     - y_pred
     - y
     - sensitiveValue

    """


    def __init__(self, y_pred=None, y=None, sensitiveValue=None,):
        self.y_pred = y_pred
        self.y = y
        self.sensitiveValue = sensitiveValue

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.y_pred = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I32:
                    self.y = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.STRING:
                    self.sensitiveValue = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Instance')
        if self.y_pred is not None:
            oprot.writeFieldBegin('y_pred', TType.I32, 1)
            oprot.writeI32(self.y_pred)
            oprot.writeFieldEnd()
        if self.y is not None:
            oprot.writeFieldBegin('y', TType.I32, 2)
            oprot.writeI32(self.y)
            oprot.writeFieldEnd()
        if self.sensitiveValue is not None:
            oprot.writeFieldBegin('sensitiveValue', TType.STRING, 3)
            oprot.writeString(self.sensitiveValue.encode('utf-8') if sys.version_info[0] == 2 else self.sensitiveValue)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(FairMetric)
FairMetric.thrift_spec = (
    None,  # 0
    (1, TType.I32, 'type', None, None, ),  # 1
    (2, TType.DOUBLE, 'value', None, None, ),  # 2
)
all_structs.append(SensitiveFeature)
SensitiveFeature.thrift_spec = (
    None,  # 0
    (1, TType.I32, 'featureIndex', None, None, ),  # 1
    (2, TType.STRING, 'name', 'UTF8', None, ),  # 2
    (3, TType.STRING, 'description', 'UTF8', None, ),  # 3
    (4, TType.I32, 'unprivPop', None, None, ),  # 4
    (5, TType.STRING, 'unprivPop_original', 'UTF8', None, ),  # 5
    (6, TType.I32, 'type', None, None, ),  # 6
)
all_structs.append(FeatureImportance)
FeatureImportance.thrift_spec = (
    None,  # 0
    (1, TType.I32, 'indexFeature', None, None, ),  # 1
    (2, TType.DOUBLE, 'importance', None, None, ),  # 2
)
all_structs.append(FIExplanation)
FIExplanation.thrift_spec = (
    None,  # 0
    (1, TType.LIST, 'listOfFeatureImportance', (TType.STRUCT, [FeatureImportance, None], False), None, ),  # 1
)
all_structs.append(ModelDefinitions)
ModelDefinitions.thrift_spec = (
    None,  # 0
    (1, TType.I32, 'numFeatures', None, None, ),  # 1
    (2, TType.LIST, 'nonNumericFeatures', (TType.I32, None, False), None, ),  # 2
    (3, TType.I32, 'type', None, None, ),  # 3
)
all_structs.append(Instance)
Instance.thrift_spec = (
    None,  # 0
    (1, TType.I32, 'y_pred', None, None, ),  # 1
    (2, TType.I32, 'y', None, None, ),  # 2
    (3, TType.STRING, 'sensitiveValue', 'UTF8', None, ),  # 3
)
fix_spec(all_structs)
del all_structs



import numpy as np

class FixOutArtifact:
    """
    A fixout artifact.

    A class for handling training and testing data.

    See Also
    --------
    fixout.helper.FixOutHelper : A FixOut helper.

    """

    def __init__(self,
                 features_name,
                 training_data,
                 testing_data=[],
                 nonnumeric_features=[],
                 model=None, 
                 y_pred=None,
                 prob_y_pred=None, 
                 sensitive_features=[], 
                 dictionary=None):
        """
        Initializes the FixOutArtifact object with the provided data and configuration.

        Parameters:
        -----------
        features_name : list
            A list of feature names used in the dataset.
        training_data : tuple or list
            A tuple or list containing two elements:
            - A 2D array or matrix (ndarray) of training features.
            - A 1D array (ndarray) of target values.
        testing_data : list of tuples, optional
            A list of tuples containing testing data. Each tuple should be of the form (X_test, y_test, label).
            If no testing data is provided, an empty list will be used (default is None).
        nonnumeric_features : list, optional
            A list of indices of non-numeric features. Default is an empty list.
        model : object, optional
            A machine learning model to be used for predictions. Default is None.
        y_pred : array-like, optional
            Predicted target values from the model. Default is None.
        prob_y_pred : array-like, optional
            Predicted probabilities for each class from the model. Default is None.
        sensitive_features : list, optional
            A list of sensitive features used for fairness evaluation. Default is an empty list.
        dictionary : dict, optional
            A dictionary for any additional information. Default is None.

        Raises:
        -------
        ValueError
            If `training_data` is not a tuple or list containing exactly two elements (features and targets).
        """
        
        self.nonnumeric_features = nonnumeric_features
        self.features_name = features_name
        
        self.model = model
        self.y_pred = y_pred
        self.prob_y_pred = prob_y_pred
          
        self.X = training_data[0]
        self.y = training_data[1]

        if self.X is not None:
            if not isinstance(self.X, (np.ndarray)):
                self.X = np.array(self.X)
                # todo check the number of lines with X_train

        if self.y is not None:
            if not isinstance(self.y, (np.ndarray)):
                self.y = np.array(self.y)
                # todo check the number of lines with X_train
        
        self.test_data = []

        for i in range(len(testing_data)):

            X,y,label = testing_data[i]

            if X is not None:
                if not isinstance(X, (np.ndarray)):
                    X = np.array(X)

            if y is not None:
                if not isinstance(self.y, (np.ndarray)):
                    y = np.array(y)

            self.test_data.append((X,y,label))

        
        self.sensfeatList = sensitive_features
        self.dictionary = dictionary

        # if ... check if all of them are not None at the same time
        
